from autosar.writer.writer_base import ElementWriter
import autosar.datatype

class SignalWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)

    def getSupportedXML(self):
        return ['SYSTEM-SIGNAL']

    def getSupportedCode(self):
        return []

    def writeElementXML(self, elem):
        if type(elem).__name__ == 'SYSTEM-SIGNAL':
            return self.writeSignalXML(elem)
        else:
            return None

    def writeElementCode(self, elem, localvars):
        raise NotImplementedError('writeElementCode')

    def writeSignalXML(self,elem):
        assert isinstance(elem,autosar.signal.SystemSignal)
        lines = []
        lines.append('<SYSTEM-SIGNAL>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        if elem.desc is not None:
            tmp = self.writeDescXML(elem)
            if tmp is not None: lines.extend(self.indent(tmp,1))
        if elem.dataTypeRef is not None:
            ws = elem.rootWS()
            typeElem = ws.find(elem.dataTypeRef, role="DataType")
            if (typeElem is None):
                raise ValueError("invalid type reference: '%s'"%elem.typeRef)
            else:
                lines.append(self.indent('<DATA-TYPE-REF DEST="%s">%s</DATA-TYPE-REF>'%(typeElem.tag(self.version),typeElem.ref),1))
        if elem.initValueRef is not None:
            ws = elem.rootWS()
            constantElem=ws.find(elem.initValueRef,role="Constant")
            if (constantElem is None): ##   start cheating patch for handle "role" for arrays & records constant refs, looking at datatype instead of Const REC element type. Shall be same for a consitant designed system.
                datatypeElem=ws.find(elem.dataTypeRef, role="DataType")
                if datatypeElem.tag(self.version) == 'INTEGER-TYPE':
                    xmlValueTypeTxt= 'INTEGER-LITERAL'
                elif datatypeElem.tag(self.version) == 'STRING-TYPE':
                    xmlValueTypeTxt= 'STRING-LITERAL'
                elif datatypeElem.tag(self.version) == 'BOOLEAN-TYPE':
                    xmlValueTypeTxt= 'BOOLEAN-LITERAL'   ## end cheating patch for record arrays
                else:
                    raise ValueError("invalid type reference: '%s'" % elem.typeRef)
            else:
                xmlValueTypeTxt=constantElem.tag(self.version)

            lines.append(self.indent('<INIT-VALUE-REF DEST="%s">%s</INIT-VALUE-REF>'%(xmlValueTypeTxt,elem.initValueRef),1))
        if elem.length is not None:
            lines.append(self.indent('<LENGTH>%s</LENGTH>'%(elem.length),1))
        lines.append('</SYSTEM-SIGNAL>')
        return lines



import cfile as C
import os
import io
import autosar

innerIndentDefault=3 #default indendation (number of spaces)

def _genCommentHeader(comment):
   code = C.sequence()
   code.append(C.line('/*********************************************************************************************************************'))
   code.append(C.line('* %s'%comment))
   code.append(C.line('*********************************************************************************************************************/'))
   return code

class AlarmVariable:
   def __init__(self, task):
      init_delay=0 #FIXME: allow users to select this at a later time
      self.decl = C.variable('os_alarm_cfg_%s'%task.name, 'os_alarm_cfg_t', static=True, const=True, array='OS_NUM_ALARMS_%s'%task.name)
      self.body = C.block(innerIndent=innerIndentDefault)
      self.body.append(C.linecomment('OS Task,       Event ID,                     Init Delay (ms),  Period (ms)'))
      for event in task.timer_events:
         self.body.append(C.line('{'+'{0: >10},{1: >50},{2: >5},{3: >5}'.format(
            '&m_os_task_'+task.name, 'EVENT_MASK_%s_%s'%(task.name,event.name), init_delay, event.inner.period)+'},'))

class OsTaskCfgVar:
   def __init__(self, tasks):
      self.decl = C.variable('os_task_cfg', 'os_task_elem_t', static=True, const=True, array='OS_NUM_TASKS')
      self.body = C.block(innerIndent=innerIndentDefault)
      for task in tasks:
         fmt='{0: >25},{1: >15},{2: >30},{3: >30}'
         if len(task.timer_events)>0:
            self.body.append(C.line('{'+fmt.format(
               '&m_os_task_'+task.name, task.name, '&os_alarm_cfg_%s[0]'%task.name, 'OS_NUM_ALARMS_%s'%task.name)+'},'))
         else:
            self.body.append(C.line('{'+fmt.format(
               '&m_os_task_'+task.name, task.name, '(os_alarm_cfg_t*) 0', '0')+'},'))

class OsConfigGenerator:
   def __init__(self, cfg):
      self.cfg = cfg
      self.static_vars={}
      self.alarm_vars=[]
      self.os_task_var=None
      
   def generate(self, dest_dir='.'):
      for task in self.cfg.tasks:
         task.finalize()
      self._create_static_vars()
      self.os_task_var = OsTaskCfgVar(self.cfg.tasks)
      self._generate_event_cfg_header(dest_dir)
      header_file = self._generate_task_cfg_header(dest_dir)
      self._generate_task_cfg_source(dest_dir, header_file)
   
   def _create_static_vars(self):
      for os_task in self.cfg.tasks:
         static_var = C.variable('m_os_task_'+os_task.name, 'os_task_t', static=True)
         self.static_vars[static_var.name]=static_var
         if len(os_task.timer_events)>0:
            self.alarm_vars.append(AlarmVariable(os_task))
   
   def _generate_event_cfg_header(self, dest_dir, file_name='os_event_cfg.h'):      
      header = C.hfile(os.path.join(dest_dir, file_name))
      code = header.code
      code.extend(_genCommentHeader('INCLUDES'))
      code.append(C.include('PlatForm_Types.h'))
      code.append('')
      code.extend(_genCommentHeader('PUBLIC CONSTANTS AND DATA TYPES'))
      for os_task in self.cfg.tasks:
         for event_mask in os_task.event_masks:
            code.append(event_mask)
         code.append(C.define('OS_NUM_ALARMS_%s'%os_task.name, str(len(os_task.timer_events))))
         code.append(C.blank())
      with io.open(header.path, 'w', newline='\n') as fp:
         for line in header.lines():
            fp.write(line+'\n')

   def _generate_task_cfg_header(self, dest_dir, file_name = 'os_task_cfg.h'):
      header = C.hfile(os.path.join(dest_dir, file_name))
      code = header.code
      code.extend(_genCommentHeader('INCLUDES'))
      code.append(C.include('os_types.h'))
      code.append(C.include('os_task.h'))
      code.append('')
      code.extend(_genCommentHeader('PUBLIC CONSTANTS AND DATA TYPES'))
      code.append(C.define('OS_NUM_TASKS',str(len(self.cfg.tasks))+'u'))
      code.append('')
      code.extend(_genCommentHeader('PUBLIC VARIABLES'))
      code.append(C.statement('extern os_cfg_t g_os_cfg'))
      code.append('')
      code.extend(_genCommentHeader('PUBLIC FUNCTION PROTOTYPES'))
      for task in self.cfg.tasks:
         code.append(C.statement('OS_TASK_HANDLER(%s, arg)'%task.name))
      for function_name in self.cfg.mode_switch_calls:
         code.append(C.statement(C.function(function_name, 'void')))
      with io.open(header.path, 'w', newline='\n') as fp:
         for line in header.lines():
            fp.write(line+'\n')
      return file_name

   def _generate_task_cfg_source(self, dest_dir, header_file, file_name = 'os_task_cfg.c'):
      source = C.cfile(os.path.join(dest_dir, file_name))
      code = source.code
      code.extend(_genCommentHeader('INCLUDES'))
      code.append(C.include(header_file))
      code.append(C.include('os_event_cfg.h'))
      code.append('')
      code.extend(_genCommentHeader('PRIVATE VARIABLES'))
      for static_var in sorted(self.static_vars.values(), key=lambda x: x.name):
         code.append(C.statement(static_var))
      code.append('')
      for alarm_var in self.alarm_vars:
         code.append(C.line(str(alarm_var.decl)+' ='))
         code.append(C.statement(alarm_var.body))
      code.append(C.line(str(self.os_task_var.decl)+' ='))
      code.append(C.statement(self.os_task_var.body))
      code.append('')
      code.extend(_genCommentHeader('PUBLIC VARIABLES'))
      code.append(C.line('os_cfg_t g_os_cfg ='))
      body = C.block(innerIndent = innerIndentDefault)
      body.append(C.line('&os_task_cfg[0],'))
      body.append(C.line('OS_NUM_TASKS,'))
      body.append(C.line('0,'))
      body.append(C.line('0'))
      code.append(C.statement(body))
      code.append('')
      code.extend(_genCommentHeader('PUBLIC FUNCTIONS'))
      for elem in self.cfg.partition.mode_switch_functions.values():
         for callback_name in sorted(elem.calls.keys()):
            code.extend(self._generate_mode_switch_func(callback_name, elem.calls[callback_name]))
            code.append('')

      with io.open(source.path, 'w', newline='\n') as fp:
         for line in source.lines():
            fp.write(line+'\n')
   
   def _generate_mode_switch_func(self, callback_name, events):
      code = C.sequence()
      generated=set()
      code.append(C.function(callback_name, 'void'))
      block = C.block(innerIndent = innerIndentDefault)
      for event in events:
         task = self.cfg.find_os_task_by_runnable(event.runnable)
         if task is not None:
            if (task.name, event.name) not in generated:               
               block.append(C.statement(C.fcall('os_task_setEvent', params=['&m_os_task_%s'%task.name, 'EVENT_MASK_%s_%s'%(task.name, event.name)])))
               generated.add((task.name, event.name))
      code.append(block)
      return code


from autosar.base import parseXMLFile,splitRef,parseTextNode,parseIntNode
from autosar.signal import *
from autosar.parser.parser_base import ElementParser

class SignalParser(ElementParser):
    def __init__(self,version=3):
        self.version=version

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {'SYSTEM-SIGNAL': self.parseSystemSignal,
                             'SYSTEM-SIGNAL-GROUP': self.parseSystemSignalGroup
            }
        elif self.version >= 4.0:
            self.switcher = {
            }

    def getSupportedTags(self):
        return self.switcher.keys()

    def parseElement(self, xmlElement, parent = None):
        parseFunc = self.switcher.get(xmlElement.tag)
        if parseFunc is not None:
            return parseFunc(xmlElement,parent)
        else:
            return None

    def parseSystemSignal(self,xmlRoot,parent=None):
        """
        parses <SYSTEM-SIGNAL>
        """
        assert(xmlRoot.tag=='SYSTEM-SIGNAL')
        name,dataTypeRef,initValueRef,length,desc=None,None,None,None,None
        for elem in xmlRoot.findall('./*'):
            if elem.tag=='SHORT-NAME':
                name=parseTextNode(elem)
            elif elem.tag=='DATA-TYPE-REF':
                dataTypeRef=parseTextNode(elem)
            elif elem.tag=='INIT-VALUE-REF':
                initValueRef=parseTextNode(elem)
            elif elem.tag=='LENGTH':
                length=parseIntNode(elem)
            elif elem.tag=='DESC':
                descXml = xmlRoot.find('DESC')
                if descXml is not None:
                    L2Xml = descXml.find('L-2')
                    if L2Xml is not None:
                        desc = parseTextNode(L2Xml)
            else:
                raise NotImplementedError(elem.tag)
#      if (name is not None) and (dataTypeRef is not None) and (initValueRef is not None) and length is not None:
        if (name is not None) and length is not None:  #All signals doesn't have IV constant Ref or DatatypeRef
            return SystemSignal(name, dataTypeRef, initValueRef, length, desc, parent)
        else:
            raise RuntimeError('failed to parse %s'%xmlRoot.tag)

    def parseSystemSignalGroup(self, xmlRoot, parent=None):
        name,systemSignalRefs=None,None
        for elem in xmlRoot.findall('./*'):
            if elem.tag=='SHORT-NAME':
                name=parseTextNode(elem)
            elif elem.tag=='SYSTEM-SIGNAL-REFS':
                systemSignalRefs=[]
                for childElem in elem.findall('./*'):
                    if childElem.tag=='SYSTEM-SIGNAL-REF':
                        systemSignalRefs.append(parseTextNode(childElem))
                    else:
                        raise NotImplementedError(childElem.tag)
            else:
                raise NotImplementedError(elem.tag)

        if (name is not None) and (isinstance(systemSignalRefs,list)):
            return SystemSignalGroup(name,systemSignalRefs)
        else:
            raise RuntimeError('failed to parse %s'%xmlRoot.tag)



from autosar.element import Element
import autosar.constant
from autosar.base import hasAdminData,parseAdminDataNode
from autosar.parser.parser_base import ElementParser

class ConstantParser(ElementParser):
    """
    Constant package parser
    """
    def __init__(self,version=3.0):
        super().__init__(version)

    def getSupportedTags(self):
        return ['CONSTANT-SPECIFICATION']

    def parseElement(self, xmlElement, parent = None):
        if xmlElement.tag == 'CONSTANT-SPECIFICATION':
            return self.parseConstantSpecification(xmlElement, parent)
        else:
            return None

    def parseConstantSpecification(self, xmlElem, rootProject=None, parent=None):
        assert(xmlElem.tag == 'CONSTANT-SPECIFICATION')
        (xmlValue, xmlValueSpec) = (None, None)
        self.push()
        for xmlElem in xmlElem.findall('./*'):
            if self.version < 4.0 and xmlElem.tag == 'VALUE':
                xmlValue = xmlElem
            elif self.version >= 4.0 and xmlElem.tag == 'VALUE-SPEC':
                xmlValueSpec = xmlElem
            elif xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            else:
                self.defaultHandler(xmlElem)

        if (self.name is not None) and ((xmlValue is not None) or (xmlValueSpec is not None)):
            constant = autosar.constant.Constant(self.name, parent=parent, adminData=self.adminData)
            if xmlValue is not None:
                constant.value = self._parseValueV3(xmlValue.find('./*') , constant)
            elif xmlValueSpec is not None:
                values = self.parseValueV4(xmlValueSpec, constant)
                if len(values) != 1:
                    raise ValueError('A value specification must contain exactly one element')
                constant.value = values[0]
            retval = constant
        else:
            retval = None
        self.pop(retval)
        return retval

    def _parseValueV3(self, xmlValue, parent):
        constantValue = None
        xmlName = xmlValue.find('SHORT-NAME')
        if xmlName is not None:
            name=xmlName.text
            if xmlValue.tag == 'INTEGER-LITERAL':
                typeRef = xmlValue.find('./TYPE-TREF').text
                innerValue = xmlValue.find('./VALUE').text
                constantValue = autosar.constant.IntegerValue(name, typeRef, innerValue, parent)
            elif xmlValue.tag=='STRING-LITERAL':
                typeRef = xmlValue.find('./TYPE-TREF').text
                innerValue = xmlValue.find('./VALUE').text
                constantValue = autosar.constant.StringValue(name, typeRef, innerValue, parent)
            elif xmlValue.tag=='BOOLEAN-LITERAL':
                typeRef = xmlValue.find('./TYPE-TREF').text
                innerValue = xmlValue.find('./VALUE').text
                constantValue = autosar.constant.BooleanValue(name, typeRef, innerValue, parent)
            elif xmlValue.tag == 'RECORD-SPECIFICATION' or xmlValue.tag == 'ARRAY-SPECIFICATION':
                typeRef = xmlValue.find('./TYPE-TREF').text
                if xmlValue.tag == 'RECORD-SPECIFICATION':
                    constantValue=autosar.constant.RecordValue(name, typeRef, parent=parent)
                else:
                    constantValue=autosar.constant.ArrayValue(name, typeRef, parent=parent)
                for innerElem in xmlValue.findall('./ELEMENTS/*'):
                    innerConstant = self._parseValueV3(innerElem, constantValue)
                    if innerConstant is not None:
                        constantValue.elements.append(innerConstant)
        return constantValue

    def parseValueV4(self, xmlValue, parent):
        result = []
        for xmlElem in xmlValue.findall('./*'):
            if xmlElem.tag == 'TEXT-VALUE-SPECIFICATION':
                result.append(self._parseTextValueSpecification(xmlElem, parent))
            elif xmlElem.tag == 'RECORD-VALUE-SPECIFICATION':
                result.append(self._parseRecordValueSpecification(xmlElem, parent))
            elif xmlElem.tag == 'NUMERICAL-VALUE-SPECIFICATION':
                result.append(self._parseNumericalValueSpecification(xmlElem, parent))
            elif xmlElem.tag == 'ARRAY-VALUE-SPECIFICATION':
                result.append(self._parseArrayValueSpecification(xmlElem, parent))
            elif xmlElem.tag == 'CONSTANT-REFERENCE':
                result.append(self._parseConstantReference(xmlElem, parent))
            elif xmlElem.tag == 'APPLICATION-VALUE-SPECIFICATION':
                result.append(self._parseApplicationValueSpecification(xmlElem, parent))
            else:
                raise NotImplementedError(xmlElem.tag)
        return result

    def _parseTextValueSpecification(self, xmlValue, parent):
        (label, value) = (None, None)
        for xmlElem in xmlValue.findall('./*'):
            if xmlElem.tag == 'SHORT-LABEL':
                label = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'VALUE':
                value = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)

        if value is not None:
            return autosar.constant.TextValue(label, value, parent)
        else:
            raise RuntimeError("Value must not be None")

    def _parseNumericalValueSpecification(self, xmlValue, parent):
        (label, value) = (None, None)
        for xmlElem in xmlValue.findall('./*'):
            if xmlElem.tag == 'SHORT-LABEL':
                label = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'VALUE':
                value = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)

        if value is not None:
            return autosar.constant.NumericalValue(label, value, parent)
        else:
            raise RuntimeError("value must not be None")

    def _parseRecordValueSpecification(self, xmlValue, parent):
        (label, xmlFields) = (None, None)
        for xmlElem in xmlValue.findall('./*'):
            if xmlElem.tag == 'SHORT-LABEL':
                label = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'FIELDS':
                xmlFields = xmlElem
            else:
                raise NotImplementedError(xmlElem.tag)

        if (xmlFields is not None):
            record = autosar.constant.RecordValueAR4(label, parent=parent)
            record.elements = self.parseValueV4(xmlFields, record)
            return record
        else:
            raise RuntimeError("<FIELDS> must not be None")

    def _parseArrayValueSpecification(self, xmlValue, parent):
        (label, xmlElements) = (None, None)
        for xmlElem in xmlValue.findall('./*'):
            if xmlElem.tag == 'SHORT-LABEL':
                label = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ELEMENTS':
                xmlElements = xmlElem
            else:
                raise NotImplementedError(xmlElem.tag)

        if (xmlElements is not None):
            array = autosar.constant.ArrayValueAR4(label, parent=parent)
            array.elements = self.parseValueV4(xmlElements, array)
            return array

        else:
            raise RuntimeError("<ELEMENTS> must not be None")

    def _parseConstantReference(self, xmlRoot, parent):
        label, constantRef = None, None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-LABEL':
                label = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'CONSTANT-REF':
                constantRef = self.parseTextNode(xmlElem)
            else:
                self.baseHandler(xmlElem)
        if constantRef is not None:
            obj = autosar.constant.ConstantReference(label, constantRef, parent, self.adminData)
            self.pop(obj)
            return obj
        else:
            raise RuntimeError('<CONSTANT-REF> must not be None')

    def _parseApplicationValueSpecification(self, xmlRoot, parent):
        label, swValueCont, swAxisCont, category = None, None, None, None

        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-LABEL':
                label = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'CATEGORY':
                category = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SW-VALUE-CONT':
                swValueCont = self._parseSwValueCont(xmlElem)
            elif xmlElem.tag == 'SW-AXIS-CONTS':
                xmlChild = xmlElem.find('./SW-AXIS-CONT')
                if xmlChild is not None:
                    swAxisCont = self._parseSwAxisCont(xmlChild)
            else:
                raise NotImplementedError(xmlElem.tag)
        value = autosar.constant.ApplicationValue(label, swValueCont = swValueCont, swAxisCont = swAxisCont, category = category, parent = parent)
        return value

    def _parseSwValueCont(self, xmlRoot):
        unitRef = None
        valueList = []
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'UNIT-REF':
                unitRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SW-VALUES-PHYS':
                for xmlChild in xmlElem.findall('./*'):
                    if (xmlChild.tag == 'V') or (xmlChild.tag == 'VF'):
                        valueList.append(self.parseNumberNode(xmlChild))
                    elif xmlChild.tag == 'VT':
                        valueList.append(self.parseTextNode(xmlChild))
                    else:
                        raise NotImplementedError(xmlChild.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        if len(valueList)==0:
            valueList = None
        return autosar.constant.SwValueCont(valueList, unitRef)

    def _parseSwAxisCont(self, xmlRoot):
        unitRef = None
        valueList = []
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'UNIT-REF':
                unitRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SW-VALUES-PHYS':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'V':
                        valueList.append(self.parseNumberNode(xmlChild))
                    else:
                        raise NotImplementedError(xmlChild.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        if len(valueList)==0:
            valueList = None
        return autosar.constant.SwAxisCont(valueList, unitRef)


from autosar.element import Element

class System(Element):
    def __init__(self,name,parent=None):
        super().__init__(name,parent)
        self.fibexElementRefs=[]
        self.mapping=None
        self.softwareComposition=None

    def asdict(self):
        data={'type': self.__class__.__name__,'name':self.name,
              }
        if len(self.fibexElementRefs)>0:
            data['fibexElementRefs']=self.fibexElementRefs[:]
        return data

class DataMapping:
    def __init__(self):
        self.swcToImpl=[]
        self.senderReceiverToSignal=[]
        self.senderReceiverToSignalGroup=[]

class Mapping:
    def __init__(self,name=None,parent=None):
        self.name=None
        self.data=DataMapping()

class SenderReceiverToSignalMapping:
    """
    <SENDER-RECEIVER-TO-SIGNAL-MAPPING>
    """
    def __init__(self,dataElemInstanceRef,signalRef):
        self.dataElemInstanceRef=dataElemInstanceRef
        self.signalRef=signalRef

class SignalDataElementInstanceRef:
    """
    <DATA-ELEMENT-IREF>
    Note: Observe that there are multiple <DATA-ELEMENT-IREF> definitions in the AUTOSAR XSD (used for different purposes)
    """
    def __init__(self,dataElemRef):
        self.dataElemRef = dataElemRef      #minOccurs=1, maxOccurs=1
        self.softwareCompositionRef=None    #minOccurs=0, maxOccurs=1
        self.componentPrototypeRef=[]       #minOccurs=0, maxOccurs=unbounded
        self.portPrototypeRef=None          #minOccurs=0, maxOccurs=1
    def asdict(self):
        data={'type': self.__class__.__name__,'dataElemRef':self.dataElemRef}
        return data

class SenderReceiverToSignalGroupMapping:
    """
    <SENDER-RECEIVER-TO-SIGNAL-GROUP-MAPPING>
    """
    def __init__(self,dataElemInstanceRef,signalGroupRef,typeMapping):
        self.dataElemInstanceRef=dataElemInstanceRef
        self.signalGroupRef=signalGroupRef
        self.typeMapping=typeMapping

class TypeMapping:
    def __init__(self):
        self.elements=[]


class ArrayElementMapping(TypeMapping):
    def __init__(self):
        super().__init__()

class RecordElementMapping(TypeMapping):
    def __init__(self):
        super().__init__()

class SenderRecRecordElementMapping:
    def __init__(self,recordElementRef,signalRef):
        self.recordElementRef=recordElementRef
        self.signalRef=signalRef



import autosar.base
import autosar.component
import autosar.rte.base
from autosar.rte.base import (ReadPortFunction, WritePortFunction, SendPortFunction, ReceivePortFunction, CallPortFunction,
                              CalPrmPortFunction, DataElement, Operation, RequirePort, ProvidePort)
import cfile as C
import sys
import autosar.bsw.com


innerIndentDefault=3 #(number of spaces)


   

class ComponentAPI:
   """
   defines the API both for clients (components) and server (RTE)
   """
   def __init__(self):
      self.read = {}
      self.write = {}
      self.send = {}
      self.receive = {}
      self.mode = {}
      self.call = {}
      self.calprm = {}
      self.get = {}
      self.setReadData = {}
      self.setReadResult = {}

      self.final = {
                     'read': [],
                     'write': [],
                     'receive': [],
                     'send': [],
                     'mode': [],
                     'call': [],
                     'calprm': [],
                     'modeswitch': [],
                     'get': [],           #FOR UNIT TEST PURPOSES
                     'setReadData': [],           #FOR UNIT TEST PURPOSES
                     'setReadResult': [],        #FOR UNIT TEST PURPOSES
                   }

   def finalize(self):
      if len(self.read)>0:
         self.final['read']=[self.read[k] for k in sorted(self.read.keys())]
      if len(self.write)>0:
         self.final['write']=[self.write[k] for k in sorted(self.write.keys())]
      if len(self.receive)>0:
         self.final['receive']=[self.receive[k] for k in sorted(self.receive.keys())]
      if len(self.mode)>0:
         self.final['receive']=[self.mode[k] for k in sorted(self.mode.keys())]
      if len(self.call)>0:
         self.final['call']=[self.call[k] for k in sorted(self.call.keys())]
      if len(self.calprm)>0:
         self.final['calprm']=[self.calprm[k] for k in sorted(self.calprm.keys())]
      if len(self.get)>0:
         self.final['get']=[self.get[k] for k in sorted(self.get.keys())]
      if len(self.setReadData)>0:
         self.final['setReadData']=[self.setReadData[k] for k in sorted(self.setReadData.keys())]
      if len(self.setReadResult)>0:
         self.final['setReadResult']=[self.setReadResult[k] for k in sorted(self.setReadResult.keys())]

   def get_all(self):
      for func in self.final['read']:
         yield func
      for func in self.final['write']:
         yield func
      for func in self.final['receive']:
         yield func
      for func in self.final['send']:
         yield func
      for func in self.final['mode']:
         yield func
      for func in self.final['call']:
         yield func
      for func in self.final['calprm']:
         yield func
      for func in self.final['get']:
         yield func
      for func in self.final['setReadData']:
         yield func
      for func in self.final['setReadResult']:
         yield func

   def update(self, other):
      self.read.update(other.read)
      self.write.update(other.write)
      self.send.update(other.send)
      self.receive.update(other.receive)
      self.mode.update(other.mode)
      self.call.update(other.call)
      self.calprm.update(other.calprm)

class Component:
   """
   RTE Container class for AUTOSAR SWC
   """
   def __init__(self, swc, parent, rte_prefix='Rte'):
      self.parent = parent
      self.name = swc.name
      self.inner = swc
      self.clientAPI = ComponentAPI() #function calls towards the RTE (stuff that the RTE must provide)
      self.events = []
      self.runnables = []
      self.data_vars = []
      self.rte_prefix = rte_prefix
      self.is_finalized = False
      self.requirePorts = []
      self.providePorts = []
      self.data_element_port_access = {}
      self.operation_port_access = {}
      ws = swc.rootWS()
      assert(ws is not None)
      self._process_ports(ws)
      self._process_runnables(ws)
      self._process_events(ws)
      
      

   def _process_ports(self, ws):
      for ar_port in self.inner.providePorts:
         self.providePorts.append(ProvidePort(ws, ar_port, self))
      for ar_port in self.inner.requirePorts:
         self.requirePorts.append(RequirePort(ws, ar_port, self))      
      
   # def pre_finalize(self, ws, type_manager):
   #    if not self.is_finalized:
   #       self._process_runnables(ws)
   #       self._process_events(ws)

   def finalize(self,ws, type_manager):
      if not self.is_finalized:      
         self._process_port_access()
         for port in self.requirePorts+self.providePorts:
            port.process_types(ws, type_manager)
            port.update_client_api(self.clientAPI)
         self.clientAPI.finalize()
         self._runnables_finalize()
      self.is_finalized=True

   def get_runnable(self, name):
      return self.rte_runnables[name]

   def find_require_port(self, name):
      for port in self.requirePorts:
         if port.name == name: return port
      raise KeyError("No port found with name "+name)

   def find_provide_port(self, name):
      for port in self.providePorts:
         if port.name == name: return port
      raise KeyError("No port found with name "+name)

   def add_event(self, rte_event):
      self.rte_events.append(rte_event)

   def _process_runnables(self, ws):
      if self.inner.behavior is not None:
         for ar_runnable in self.inner.behavior.runnables:
            runnable = Runnable(self, ar_runnable)
            self.runnables.append(runnable)
            for dataPoint in ar_runnable.dataReceivePoints+ar_runnable.dataSendPoints:
               ar_port=ws.find(dataPoint.portRef)
               if ar_port is None:
                  raise ValueError('Error: Invalid port reference: '+dataPoint.dataPoint.portRef)
               ar_data_element = ws.find(dataPoint.dataElemRef)
               if ar_data_element is None:
                  raise ValueError('Error: Invalid data element reference: '+dataPoint.dataElemRef)
               if isinstance(dataPoint, autosar.behavior.DataSendPoint):
                  port = self.find_provide_port(ar_port.name)
               else:
                  port = self.find_require_port(ar_port.name)
               data_element = port.find_data_element(ar_data_element.name)
               runnable.data_element_access.append(data_element)
               port.create_data_access_api(ws, data_element)               
               self.data_element_port_access['%s/%s'%(port.name, data_element.name)]=autosar.rte.base.DataElementPortAccess(port, data_element, runnable)

            for callPoint in ar_runnable.serverCallPoints:
               for instanceRef in callPoint.operationInstanceRefs:
                  ar_port = ws.find(instanceRef.portRef)
                  if ar_port is None:
                     raise ValueError('Error: Invalid port reference: '+instanceRef.portRef)
                  ar_operation = ws.find(instanceRef.operationRef)
                  if ar_operation is None:
                     raise ValueError('Error: Invalid operation reference: '+instanceRef.operationRef)
                  port = self.find_require_port(ar_port.name)
                  operation = port.find_operation(ar_operation.name)
                  runnable.operation_access.append(operation)                  
                  self.operation_port_access['%s/%s'%(port.name, operation.name)]=autosar.rte.base.OperationPortAccess(port, operation, runnable)

   def _process_events(self, ws):
      if self.inner.behavior is None:
         return
      for ar_event in self.inner.behavior.events:
         ar_runnable = ws.find(ar_event.startOnEventRef)
         if ar_runnable is None:
            raise ValueError('Invalid StartOnEvent reference: '+ar_event.startOnEventRef)
         for runnable in self.runnables:
            if runnable.inner is ar_runnable:
               break
         else:
            raise ValueError('Runnable not found')
         if isinstance(ar_event, autosar.behavior.TimingEvent):
            event = autosar.rte.base.TimerEvent(ar_event, runnable)
         elif isinstance(ar_event, autosar.behavior.ModeSwitchEvent):
            if ar_event.modeInstRef is not None:
               event = autosar.rte.base.ModeSwitchEvent(ws, ar_event, runnable)
         elif isinstance(ar_event, autosar.behavior.OperationInvokedEvent):
            port_refs = autosar.base.splitRef(ar_event.operationInstanceRef.portRef)
            operation_refs = autosar.base.splitRef(ar_event.operationInstanceRef.operationRef)
            port = self.find_provide_port(port_refs[-1])
            assert (port is not None) and (port.ar_port is ws.find(ar_event.operationInstanceRef.portRef))
            operation = port.find_operation(operation_refs[-1])
            assert (operation is not None)
            event = autosar.rte.base.OperationInvokedEvent(ar_event, runnable, port, operation)
         else:
            raise NotImplementedError(str(type(event)))
         self.events.append(event)
            
   def _process_port_access(self):
      for access in self.operation_port_access.values():
         if isinstance(access, autosar.rte.base.OperationPortAccess):
            proto = access.port.create_server_call_api(access.operation)
         else:
            raise NotImplementedError(str(type(access)))
   def _runnables_finalize(self):
      #operation_invoke_events = [event for event in self.events if isinstance(event, OperationInvokedEvent)]         
      for runnable in self.runnables:
         if runnable.prototype is None:
            runnable.prototype = (C.function(runnable.symbol, 'void'))

   def create_data_elements(self, data_element_map):
      for provide_port in self.providePorts:
         for require_port in provide_port.connectors:
            if len(require_port.data_elements)>0:
               for port_func in require_port.portAPI.values():
                  if isinstance(port_func, (ReadPortFunction, ReceivePortFunction)) and port_func.data_element.parent is require_port:
                     data_element = provide_port.find_data_element(port_func.data_element.name)
                     assert(data_element is not None)
                     variable_name = '_'.join([self.name,provide_port.name,data_element.name])
                     if variable_name not in data_element_map:
                        data_element.symbol = variable_name
                        data_element_map[variable_name] = data_element
                        #reassign require_port data element to access the data element from the provide port
                        port_func.data_element = data_element



class Runnable:
   """RTE Runnable"""
   def __init__(self, parent, ar_runnable):
      self.parent = parent
      self.inner = ar_runnable
      self.name = ar_runnable.name
      self.symbol = ar_runnable.symbol
      self.data_element_access=[]
      self.operation_access=[]
      self.prototype = None
      self.event_triggers=[]
      self.processed=False

class Partition:

   def __init__(self, mode='full', prefix='Rte'):
      self.prefix=prefix
      self.components = [] #clients (components)
      self.upperLayerAPI = ComponentAPI() #functions that the RTE must support towards its clients
      self.lowerLayerAPI = {}
      self.types = autosar.rte.RteTypeManager() #centralized type manager
      self.isFinalized = False
      self.ws = None
      self.assemblyConnectorMap = {}
      self.data_element_map = {}
      self.mode_switch_functions = {}
      self.static_vars = {}


   def addComponent(self, swc, runnables = None, name=None):
      """
      adds software component to partition.
      Optional parameters:
      name: Can be used to override name of swc. Default is to use name from swc.
      """
      swc_name = name if name is not None else swc.name
      if isinstance(swc, (autosar.component.AtomicSoftwareComponent, autosar.bsw.com.ComComponent)):
         ws = swc.rootWS()
         assert(ws is not None)
         if self.ws is None:
            self.ws = ws
         else:
            if self.ws is not ws:
               raise ValueError('Cannot add components from different workspaces!')
         component = Component(swc, self)
         self.components.append(component)
      else:
         print("Unsupported component type: "+str(type(swc)), file=sys.stderr)


   def finalize(self):
      if not self.isFinalized:
#         for component in self.components:            
#            component.pre_finalize(self.ws, self.types)
         for component in self.components:
            component.finalize(self.ws, self.types)
            self.upperLayerAPI.update(component.clientAPI)
         for component in self.components:
            component.create_data_elements(self.data_element_map)
         self._generate_com_access()
         self.upperLayerAPI.finalize()
         for component in self.components:
            self._process_mode_switch_events(component)
      self.isFinalized=True

   def createConnector(self, portRef1, portRef2):
      """
      creates a connector between two ports
      """
      assert (self.ws is not None)
      port1  = self._analyzePortRef(portRef1)
      port2 = self._analyzePortRef(portRef2)

      providePort=None
      requirePort=None
      if isinstance(port1, RequirePort) and isinstance(port2, ProvidePort):
         requirePort, providePort = port1, port2
      elif isinstance(port1, ProvidePort) and isinstance(port2, RequirePort):
         providePort, requirePort = port1, port2
      elif isinstance(port1, RequirePort) and isinstance(port2, RequirePort):
         raise ValueError('cannot create assembly connector between two require ports')
      else:
         raise ValueError('cannot create assembly connector between two provide ports')
      self._createConnectorInternal(providePort, requirePort)

   def autoConnect(self):
      """
      Attemts to create compatible connectors between components
      """
      require_port_list = [] #list of RequirePort
      provide_port_list = [] #list of ProvidePort
      for rte_comp in self.components:
         for rte_port in rte_comp.requirePorts:
            require_port_list.append(rte_port)
         for rte_port in rte_comp.providePorts:
            provide_port_list.append(rte_port)

      for require_port in require_port_list:
         provide_port = self._findCompatibleProvidePort(require_port, provide_port_list)
         if provide_port is not None:
            self._createConnectorInternal(provide_port, require_port)

   def unconnectedPorts(self):
      """
      Returns a generator that yields all unconnected ports of this partition
      """
      for component in self.components:
         for port in component.requirePorts+component.providePorts:
            if len(port.connectors)==0:
               yield port


   def _findCompatibleProvidePort(self, require_port, provide_port_list):
      require_port_interface = self.ws.find(require_port.ar_port.portInterfaceRef)
      if require_port_interface is None: raise ValueError("Invalid port interface ref: %s"%require_port.ar_port.portInterfaceRef)
      for provide_port in provide_port_list:
         provide_port_interface = self.ws.find(provide_port.ar_port.portInterfaceRef)
         if provide_port_interface is None: raise ValueError("Invalid port interface ref: %s"%provide_port.ar_port.portInterfaceRef)
         if require_port_interface==provide_port_interface and (require_port.ar_port.name == provide_port.ar_port.name):
            return provide_port
      return None


   def _createConnectorInternal(self, provide_port, require_port):
      connectorName='_'.join([provide_port.parent.name, provide_port.name, require_port.parent.name, require_port.name])
      if connectorName in self.assemblyConnectorMap:
         raise ValueError('connector "%s" already exists'%connectorName)
      self.assemblyConnectorMap[connectorName]=(provide_port,require_port)
      provide_port.connectors.append(require_port)
      require_port.connectors.append(provide_port)

   # def _process_parameter_ports(self, component, ws, swc):
   #    for port in swc.requirePorts:
   #       portInterface = ws.find(port.portInterfaceRef)
   #       if portInterface is not None:
   #          if isinstance(portInterface, autosar.portinterface.ParameterInterface):
   #             for data_element in portInterface.dataElements:
   #                data_type = ws.find(data_element.typeRef)
   #                if data_type is None:
   #                   raise ValueError("Error: Invalid type reference: "+ data_element.typeRef)
   #                self.types.processType(ws, data_type)
   #                component.create_parameter(ws, port, data_element, data_type)

   # def _process_events(self, component, ws, swc, runnables=None):
   #    for event in swc.behavior.events:
   #       ar_runnable = ws.find(event.startOnEventRef)
   #       rte_runnable = component.get_runnable(ar_runnable.name)
   #       if isinstance(event, autosar.behavior.TimingEvent):
   #          rte_event = TimerEvent(event.name, rte_runnable)
   #          component.add_event(rte_event)
   #       elif isinstance(event, autosar.behavior.ModeSwitchEvent):
   #          if ar_runnable is None:
   #             raise ValueError('invalid reference: '+event.startOnEventRef)
   #          if runnables is None or ar_runnable.name in runnables:
   #             rte_event = ModeSwitchEvent(ws, event, rte_runnable)
   #             rte_runnable.events.append(rte_event)
   #             component.add_event(rte_event)
   #       elif isinstance(event, autosar.behavior.OperationInvokedEvent):
   #          pass #already processed
   #       else:
   #          raise NotImplementedError(type(event))

   def _analyzePortRef(self, portRef):
      parts=autosar.base.splitRef(portRef)
      if len(parts)==2:
         #assume format 'componentName/portName' with ComponentType role set
         port=None
         for component in self.components:
            if component.name == parts[0]:
               for port in component.requirePorts + component.providePorts:
                  if parts[1] == port.name:
                     return port
      return None

   def _generate_com_access(self):
      for component in self.components:
         if isinstance(component.inner, autosar.bsw.com.ComComponent):
            for port in component.requirePorts:
               for remote_port in port.connectors:
                  for data_element in remote_port.data_elements:
                     isPointer = True if data_element.dataType.isComplexType else False
                     proto = C.function(
                     "%s_Send_%s_%s"%(component.inner.name, remote_port.name, data_element.name),
                     'Std_ReturnType',
                     args = [C.variable('value', data_element.dataType.name, pointer=isPointer)])
                     data_element.com_access['Send'] = proto
                     component.inner.addSendInterface(proto, port, data_element)
            for port in component.providePorts:
               for data_element in port.data_elements:
                  isPointer = True
                  proto = C.function(
                  "%s_Receive_%s_%s"%(component.inner.name, remote_port.name, data_element.name),
                  'Std_ReturnType',
                  args = [C.variable('value', data_element.dataType.name, pointer=isPointer)])
                  data_element.com_access['Receive'] = proto
                  component.inner.addReceiveInterface(proto, port, data_element)
                  #remove from internal RTE variables
                  symbol = data_element.symbol
                  data_element.symbol = None
                  if symbol in self.data_element_map:
                     del self.data_element_map[symbol]
   
   def _process_mode_switch_events(self, component):      
      for event in component.events:
         if isinstance(event, autosar.rte.base.ModeSwitchEvent):
            if event.mode not in self.mode_switch_functions:
               func = autosar.rte.base.ModeSwitchFunction(event)
               self.mode_switch_functions[event.mode] = func
               if func.static_var not in self.static_vars:
                  self.static_vars[func.static_var.name] = func.static_var
            function_name = "_".join(['os', 'task', event.activationType, event.mode, event.modeDeclaration])
            if function_name not in self.mode_switch_functions[event.mode].calls:
               if (event.activationType == 'OnEntry'):
                  self.mode_switch_functions[event.mode].generate_on_entry_code(event, function_name)
               else:
                  self.mode_switch_functions[event.mode].generate_on_exit_code(event, function_name)
            else:
               self.mode_switch_functions[event.mode].add_event_to_call(event, function_name)